clear; close all
%% Set up the domain data structures
n = 500;
x_a = 0; x_b = 10; % The beginning and end of the Cartesian grid
h = (x_b - x_a)/(n-1);

m =  4; % Gram Polynomial max order
d =  5; % Number of orthogonal polynomial matching points
C = 25; % Number of continuation points
load(['FC_data_d', num2str(d), '.mat']);

% The grid we have data on and wish to interpolate/compute derivatives/solve DEs
x = (x_a:h:x_b).';
% The grid of continuation points (for plotting purposes)
x_cont = x_b + h*(1:C).';

% Period of our FC interpolating Fourier Series -- note the extended period
fourPts = n + C;
prd = fourPts*h;
if (mod(fourPts, 2) == 0)
    k = transpose([0:fourPts/2, -fourPts/2+1:-1]);
else
    k = transpose([0:(fourPts-1)/2, -(fourPts-1)/2:-1]);
end

%% Set up the initial condition of the explicit ODE
% Used for shifting the initial condition relative to the solution domain
shift = 0;

% Gaussian-modulated Wave Packet
%sigma = sqrt(1/12.5);
%omega = 15; % = 0 for a simple gaussian
%uexact = @(x, t) exp(-(x - t).^2 / sigma^2) .* cos(omega*(x - t));
%uexact_der = @(x, t) -2/sigma^2*(x - t).*uexact(x, t) - ...
%    omega * exp(-(x - t)/sigma^2).*sin(omega*(x - t));

% High frequency wave
kappa = 5;
uexact = @(x, t) sin(kappa*(x - t));
uexact_der = @(x, t) kappa*cos(kappa*(x - t));

uexact = @(x, t) uexact(x, t + shift);
uexact_der = @(x, t) uexact_der(x, t + shift);
u0x = uexact(x, 0);

% TODO: consider deleting this
% For plotting the FC representation of uexact at some time ti
%ti = 9.136;
%ux = uexact(x, ti);
%ux_fc_coeff = fcont_gram_blend(ux, d, C, Q, A).';
%[~,ux_fc_coeff] = fc_der(ux, 1, 1, prd, k, d, C, Q, A);
%x_fine = x_a:h/20:x_b;
%ux_fc = real(ux_fc_coeff * exp(2*pi*1j*k/prd * (x_fine - x_a)));
%plot(x_fine, uexact(x_fine, ti) - ux_fc);
%plot(x_fine, uexact_der(x_fine, ti) - ux_fc);
%return;

tic;
deltat = 10^-4*50/3;
n_ts = 15000;
filter = 0; % Boolean for spectral filter
sfreq = 50; % Timestep frequency for plotting solution
f = @(t, y) -fc_der(y, 1, filter, prd, k, d, C, Q, A);

errvec = ones(n_ts, 1)*NaN;
u0 = uexact(x, -3*deltat);
u1 = uexact(x, -2*deltat);
u2 = uexact(x, -deltat);
u  = uexact(x, 0);
fprintf('Performing %d timesteps to t = %1.2e\n', n_ts, n_ts*deltat);
figure(1)
for ts=1:n_ts
    % Forward Euler method
%     u = u + deltat*f(ts*deltat, u);
%     u(1) = uexact(x(1), ts*deltat);

    % Adams-Bashforth 4-step method
    tmp = u;
    u = u + deltat/24 * ...
        ( -9*f(0, u0) + 37*f(0, u1) - 59*f(0, u2) + 55*f(0, u) );
    u(1) = uexact(x(1), ts*deltat);
    u0 = u1;
    u1 = u2;
    u2 = tmp;

    errvec(ts) = max(abs(u - uexact(x, ts*deltat)));

    if (mod(ts, sfreq) == 0)
        subplot(2,2, [1,2])
        hold off
        plot(x, u0x,  'b-'); hold on
        plot(x, u, 'r-.'); legend('u(x, 0)', 'u(x, t)');
        xlabel('x', 'interpreter', 'latex', 'FontSize', 15);
        ylabel('u(x)', 'interpreter', 'latex', 'FontSize', 15);
        set(gca, 'yLim', [-1.2 1.2]);
        title(['Solution $t = ', num2str(ts*deltat), '$'], 'interpreter', 'latex');

        subplot(2, 2, 3)
        plottime = ts*deltat;
        err = u - uexact(x, plottime);
        semilogy(x, abs(err));
        set(gca, 'YLim', [10^-20 1]);
        xlabel('x', 'interpreter', 'latex', 'FontSize', 15);
        title(['Error at $t = ', num2str(plottime), '$'], 'interpreter', 'latex');

        subplot(2, 2, 4)
        semilogy( (1:n_ts)*deltat, errvec);
        set(gca, 'XLim', [0 n_ts*deltat])
        xlabel('t', 'interpreter', 'latex', 'FontSize', 15);
        title('Time Evolution of $L^\infty$ Error', 'interpreter', 'latex')
        drawnow; M(:, ts/sfreq) = getframe;
    end
end
toc;
fprintf('Final L-inf error: %1.3e\n', max(abs(err)));
