function [u,flag,relres,iter] = fc_ode_gmres(p, q, f, B_l, B_r, gmresParam, h, k, d, C, Q, A)

% FC_ODE_GMRES - Routine to compute the solution of the ODE boundary value problem
% v(x) - p(x)*v'(x) - q(x)*v''(x) = f(x); v(x_l) = B_l, v(x_r) = B_r
%
%   Inputs:
%       p - Coefficient function evaluated at grid points (real n-vector)
%       q - Coefficient function evaluated at grid points (real n-vector)
%       f - value of the function f at the nodes x_cart (real n-vector)
%       |-> h - spatial discretization size (real) (Note: h = 1/(n-1))
%       |-> x_l - left boundary point (boundary values given at these pts)
%       |-> x_r - right boundary point
%       |-> cntr - structure containing continuation data and parameters
%           |-> n_delta - no of boundary section points (integer)
%           |-> n_d - no. of continuation points (integer)
%           |-> m - maximum order of Gram polynomial (integer)
%           |-> prd - period of the FC(Gram) approximation  (real)
%           |-> k - Fourier modes of FC(Gram) approximation (integer vec)
%       B_l - left boundary value of the solution (real)
%       B_r - right boundary value of the solution (real)
%
%   Outputs:
%       u - Solution of the ODE at GMRES convergence (or last iteration).
%             (real, n-vector)
%       flag - boolean to indicate GMRES convergence.
%       relres - relative residual
%       iter - Number of iterations required to reach convergence
%
%   Author(s):
%       Thomas Anderson
%       Email: tanderson@caltech.edu

    % First, access data structure members so we do not need to reference
    % the structure as much, making things simpler to read
    n = length(f);

    prd = ((n+2) + C)*h;

    % The right-hand-side vector consists of B.C.'s and the function f_xj values
    rhs = zeros(n+2,1);
    rhs(1) = B_l; rhs(n+2) = B_r;
    for i=1:n
        rhs(i + 1) = f(i);
    end

    [L,U] = FD_precond_init(p, q, h);

    forward_map_handle = @(v) forward_map(p, q, v, prd, k, d, C, Q, A);
    precond_handle = @(v) FD_precond(p, q, v, h, L, U);
    precond = 1; % Change to 2 to use the FD preconditioner, or 0 for no preconditioning
    if precond == 1
        [u,flag,relres,iter] = gmres(forward_map_handle, rhs, gmresParam.rstrt, ...
            gmresParam.tol, gmresParam.maxit, precond_handle);
    elseif precond == 2
        u = FD_precond(p, q, rhs, h, L, U);
        flag = 0; relres = 0; iter=[1,1]; %dummy convergence info when only using preconditioner.
    else
        u = gmres(forward_map_handle, rhs, gmresParam.rstrt, gmresParam.tol, gmresParam.maxit);
    end
    u = u(2:end-1);
return

function u = FD_precond(p, q, f, h, L, U)
    B_l = f(1);
    B_r = f(end);

    N = length(p);
    f = f(2:end-1);

    b = zeros(N,1);

    % For u(x_1)
    b(1) = f(1) + p(1) * ( -1/(2*h) ) * B_l + ...
                q(1) * ( 1/h^2 ) * B_l;
    % For u(x_i) 2<=i<=N-1
    for i = 2:N-1
        b(i) = f(i);
    end
    % For u(x_(N+1))
    b(N) = f(N) + p(N) * ( 1/(2*h) ) * B_r + ...
                q(N) * ( 1/h^2 ) * B_r;

    % Use the LU decomposition to solve the system LUz = b
    y = L\b;
    z = U\y;

    u = [B_l; z; B_r];
return

function [L,U] = FD_precond_init(p, q, h)
    N = length(p);

    A = zeros(N);
    % For u(x_1)
    A(1,1) = 1 - q(1) * ( -2/h^2 );
    A(1,2) =    -p(1) * ( 1/(2*h) ) - q(1) * ( 1/h^2 );
    % For u(x_i) 2<=i<=N-1
    for i = 2:N-1
        A(i,i-1) =  p(i)/(2*h) - q(i)/h^2;
        A(i,i) = 1 + 2*q(i)/h^2;
        A(i,i+1) = -p(i)/(2*h) - q(i)/h^2;
    end
    % For u(x_N)
    A(N, N - 1) = -p(N) * ( -1/(2*h) ) - q(N) * ( 1/h^2 );
    A(N, N) = 1 - q(N) * ( -2/h^2 );

    [L,U] = lu(A);
return

function u = forward_map(p, q, v, prd, k, d, c, Q, A)
    n = length(v)-2;

    % We need both derivatives, for the ODE BVP operator
    filter = 0; % filtering is unneeded for solving a BVP
    [v_der ~] = fc_der(v, [1, 2], filter, prd, k, d, c, Q, A);
    % We solve the ODE on the cartesian grid, so ignore v^(k)(x_l), v^(k)(x_r)
    v_der = v_der(2:end-1, :);

    u = zeros(n+2, 1);
    u(1) = v(1);
    for i=2:n+1
        % Apply u <- u - p*u' - q*u''
        % v_der(:,*), p, q are of size n, v of size n+2
        u(i) = v(i) - p(i-1)*v_der(i-1,1) - q(i-1) * v_der(i-1, 2);
    end
    u(n+2) = v(end);

return

